'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _create = require('babel-runtime/core-js/object/create');

var _create2 = _interopRequireDefault(_create);

exports.default = memoize;

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _hiddenData = require('../hiddenData');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function memoize(target, property, descriptor) {
  var cacheProp = 'memoized' + property.charAt(0).toUpperCase() + property.substring(1);
  var impl = descriptor.value;

  var getHiddenData = (0, _hiddenData.createHiddenDataGetter)(cacheProp);
  var setHiddenData = (0, _hiddenData.createHiddenDataSetter)(cacheProp);

  if (impl.length === 0) {
    descriptor.value = function () {
      var value = getHiddenData(this);

      if (typeof value === 'undefined') {
        value = impl.call(this);
        setHiddenData(this, value);
      }

      return value;
    };
  } else {
    descriptor.value = function (input) {
      var cache = getHiddenData(this);

      if (typeof cache === 'undefined') {
        cache = (0, _create2.default)(null);
        setHiddenData(this, cache);
      }

      if (input in cache) {
        return cache[input];
      } else {
        var value = impl.call(this, input);
        cache[input] = value;
        return value;
      }
    };
  }
}